// A fan game by poret. This game is not affiliated with SNK. This is a fan game which is not being sold or distributed for profit.


#include "Character/KOFBaseCharacter.h"
#include "Character/KOFBaseCharacter.h"

#include "Components/InputComponent.h"
#include "Game/KOFDefaultGameMode.h"

// Sets default values
AKOFBaseCharacter::AKOFBaseCharacter()
{
 	// Set this actor to call Tick() every frame.  You can turn this off to improve performance if you don't need it.
	PrimaryActorTick.bCanEverTick = true;
	
	Sprite = CreateOptionalDefaultSubobject<UPaperFlipbookComponent>(TEXT("Sprite"));
	if (Sprite)
	{
		Sprite->AlwaysLoadOnClient = true;
		Sprite->AlwaysLoadOnServer = true;
		Sprite->bOwnerNoSee = false;
		Sprite->bAffectDynamicIndirectLighting = true;
		Sprite->SetupAttachment(RootComponent);
		Sprite->SetRelativeLocation(FVector(0.0f, 0.0f, 96.0f));
	}
	
	// Try to create the sprite component
	Shadow = CreateOptionalDefaultSubobject<UPaperFlipbookComponent>(TEXT("Shadow"));
	if (Shadow)
	{
		Shadow->AlwaysLoadOnClient = true;
		Shadow->AlwaysLoadOnServer = true;
		Shadow->bOwnerNoSee = false;
		Shadow->bAffectDynamicIndirectLighting = true;
		Shadow->SetupAttachment(Sprite);
		Shadow->SetRelativeLocation(FVector(0.0f, 0.0f, 5.0f));
		Shadow->SetRelativeRotation(FRotator(90.0f, 0.0f, 0.0f));
		Shadow->SetRelativeScale3D(FVector(1.0f, 1.0f, 1.0f));
		Shadow->SetSpriteColor(FColor(0.0f, 0.0f, 0.0f, 128));
	}
	
	MaxWalkSpeed = 450;
	GravityScale = FFixed(4);
	VerticalNormalJumpVelocity = 1550;
	HorizontalNormalJumpVelocity = MaxWalkSpeed;
	/*
	VerticalSuperJumpVelocity = 2000;
	VerticalNormalHopVelocity = 1000;
	VerticalSuperHopVelocity;
	HorizontalSuperJumpVelocity = 700;
	HorizontalNormalHopVelocity;
	HorizontalSuperHopVelocity;
	*/
	bIsInAir = false;
	bIsPossessedByPlayer1;
}

void AKOFBaseCharacter::Init()
{
	Position = FFixedVector(0, bIsPossessedByPlayer1 ? -GameMode->GetStartY() : GameMode->GetStartY(), 0);
	SetActorLocation(FVector(Position));
	// eventually character will have a forward and actor rotation will be derived from that
	SetActorRotation(FRotator(0.0f, bIsPossessedByPlayer1 ? 90.0f : -90.0f, 0.0f));
	//Shadow->SetRelativeRotation(FRotator(bIsPossessedByPlayer1 ? 90.0f : -90.0f, 0.0f, 0.0f));
}

UPaperFlipbook* AKOFBaseCharacter::GetAnimation(FName Key)
{
	return Animations.Contains(Key) ? Animations[Key] : nullptr;
}

void AKOFBaseCharacter::UpdateAnimation()
{
	// need to get the animation based on what state we're in...
	// rendering
	if(!bIsInAir)
	{
		if(bWantsToJump)
		{
			SetAnimationFlipbook("PRE_JUMP");
		}
		else
		{
			if(Direction > 0)
			{
				SetAnimationFlipbook("WALK_FWD");
			}
			else if (Direction < 0)
			{
				SetAnimationFlipbook("WALK_BACK");
			}
			else
			{
				SetAnimationFlipbook("IDLE");
			}
		}
	}
	else
	{
		SetAnimationFlipbook("JUMP_NEUTRAL");
	}
}

void AKOFBaseCharacter::SetAnimationFlipbook(FName AnimName)
{
	UPaperFlipbook* Anim = GetAnimation(AnimName);
	if(Anim)
	{
		Sprite->SetFlipbook(Anim);
		Shadow->SetFlipbook(Anim);
	}
	else
	{
		UE_LOG(LogTemp, Error, TEXT("AKOFBaseCharacter::SetAnimationFlipbook - tried to set character animation to %s but the flipbook has not been set for this state."), *AnimName.ToString());
	}
}

void AKOFBaseCharacter::SetupPlayerInputComponent(UInputComponent* PlayerInputComponent)
{
	Super::SetupPlayerInputComponent(PlayerInputComponent);

	PlayerInputComponent->BindAxis("MoveRight", this, &AKOFBaseCharacter::MoveRight);
	
	RegisterInput("Up", EVirtualGamePadButton::VGP_Up);
	RegisterInput("Down", EVirtualGamePadButton::VGP_Down);
	RegisterInput("Left", EVirtualGamePadButton::VGP_Left);
	RegisterInput("Right", EVirtualGamePadButton::VGP_Right);
}

void AKOFBaseCharacter::ProcessInputs(EVirtualGamePadButton Button)
{
	switch(Button)
	{
		case EVirtualGamePadButton::VGP_Up:
			bWantsToJump = true;
			break;
	}
}

void AKOFBaseCharacter::RegisterInput(FName ActionName, EVirtualGamePadButton Button)
{
	check(InputComponent != nullptr);
	
	FInputActionHandlerSignature ActionHandler;
	ActionHandler.BindUObject(this, &AKOFBaseCharacter::ProcessInputs, Button);

	FInputActionBinding ActionBinding(ActionName, IE_Pressed);
	ActionBinding.ActionDelegate = ActionHandler;
	InputComponent->AddActionBinding(ActionBinding);
}

void AKOFBaseCharacter::MoveRight(float Axis)
{
	Direction = Axis;
}

void AKOFBaseCharacter::UpdateMovement(FFixed DeltaTime)
{
	// should have game mode
	check(GameMode);

	if(!bIsInAir) // apply walking movement
	{
		Velocity.Y = Direction * MaxWalkSpeed;
	}
	else // apply gravity
	{
		// I keep getting confused with Z being vertical and Y being horizontal
		Velocity.Z -= (GameMode->GetGravity() * GravityScale) * DeltaTime;
	}
	
	Position +=  Velocity * DeltaTime;
	// clamp
	FFixed AdjustedZ = FFixedMath::Clamp(Position.Z, 0, 10000);
	FFixed HalfStageLength = GameMode->GetStageLength()/2;
	FFixed AdjustedY = FFixedMath::Clamp(Position.Y, -HalfStageLength, HalfStageLength);
	Position = FFixedVector(Position.X, AdjustedY, AdjustedZ);
	
	if(Position.Z <= 0) // hmmmmmmmmmmmmmmm
	{
		bIsInAir = false;
	}
}

void AKOFBaseCharacter::Jump()
{
	// can't jump if already airborne
	if(bIsInAir)
	{
		return;
	}
	
	Velocity = FFixedVector(0, HorizontalNormalJumpVelocity * Direction, VerticalNormalJumpVelocity);
	bIsInAir = true;
}

// Called when the game starts or when spawned
void AKOFBaseCharacter::BeginPlay()
{
	Super::BeginPlay();
	
	// NOTE(kevin): This will only return the game mode if we are the server. Since this game is local only we should always
	// be both server and client
	GameMode = Cast<AKOFDefaultGameMode>(GetWorld()->GetAuthGameMode());
	check(GameMode);
}

// Called every frame
void AKOFBaseCharacter::Tick(float DeltaTime)
{
	Super::Tick(DeltaTime);

	if(!bIsInAir && bWantsToJump)
	{
		TicksSinceJumpRequested++;
		if(TicksSinceJumpRequested >= 4)
		{
			Jump();
			TicksSinceJumpRequested = 0;
			bWantsToJump = false;
		}
	}
	// character logic
	UpdateMovement(FFixed(DeltaTime));
	
	SetActorLocation(FVector(Position));
	Shadow->SetWorldLocation(FVector(float(Position.Z), float(Position.Y), 5.0f));
	UpdateAnimation();
}

