// A fan game by poret. This game is not affiliated with SNK. This is a fan game which is not being sold or distributed for profit.

#pragma once

#include "CoreMinimal.h"
#include "PaperFlipbookComponent.h"
#include "../../../../Plugins/FixedPointMath/Source/FixedPointMath/Public/FixedVector.h"
#include "GameFramework/Actor.h"
#include "KOFForever/KOFTypes.h"
#include "KOFBaseCharacter.generated.h"

class UPaperFlipbook;

class AKOFDefaultGameMode;

UCLASS()
class KOFFOREVER_API AKOFBaseCharacter : public APawn
{
	GENERATED_BODY()
	
public:	
	// Sets default values for this actor's properties
	AKOFBaseCharacter();

	FORCEINLINE bool IsPossessedByPlayer1() const { return bIsPossessedByPlayer1; }

	FORCEINLINE void SetIsPossessedByPlayer1(bool bIsPlayer1) { bIsPossessedByPlayer1 = bIsPlayer1; }

	void Init();
	
protected:
	// sprite
	/** The main flipbook representing the character */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category = Character)
	UPaperFlipbookComponent* Sprite;
	
	/**
	 *	Secondary sprite that is mirrored across the ground axis and set to black to give the effect of a
	 *  character's shadow
	 */
	UPROPERTY(VisibleAnywhere, BlueprintReadOnly, Category = Character)
	UPaperFlipbookComponent* Shadow;

	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Animations")
	TMap<FName, UPaperFlipbook*> Animations;

	/**
	 * @brief Gets a character animation given a state name (i.e. WALKING, IDLE, etc.)
	 * @param Key Name of the animation
	 * @return the PaperFlipbook animation or a nullptr if the key does not exist in the animation map
	 */
	UPaperFlipbook* GetAnimation(FName Key);
	
	void UpdateAnimation();

	void SetAnimationFlipbook(FName AnimName);
	
	// movement
	// need a way to expose this to the editor as a float slider/entry
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement")
	FFixed MaxWalkSpeed;

	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement")
	FFixed GravityScale;

	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement|Jumping")
	FFixed VerticalNormalJumpVelocity;

	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement|Jumping")
	FFixed VerticalSuperJumpVelocity;
	
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement|Jumping")
	FFixed VerticalNormalHopVelocity;

	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement|Jumping")
	FFixed VerticalSuperHopVelocity;
	
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement|Jumping")
	FFixed HorizontalNormalJumpVelocity;

	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement|Jumping")
	FFixed HorizontalSuperJumpVelocity;
	
	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement|Jumping")
	FFixed HorizontalNormalHopVelocity;

	UPROPERTY(EditDefaultsOnly, BlueprintReadOnly, Category="Settings|Movement|Jumping")
	FFixed HorizontalSuperHopVelocity;
	
	int Direction;
	
	FFixedVector Position;

	FFixedVector Velocity;

	void MoveRight(float Axis);
	
	void UpdateMovement(FFixed DeltaTime);
	
	bool bIsInAir;

	bool bWantsToJump;
	int TicksSinceJumpRequested;
	
	void Jump();
	
	// input and control
	/**
	 * 
	 * @param  PlayerInputComponent
	 */
	virtual void SetupPlayerInputComponent(UInputComponent* PlayerInputComponent) override;

	// for now just takes a button but eventually will take an entire gamepad state
	void ProcessInputs(EVirtualGamePadButton Button);

	void RegisterInput(FName ActionName, EVirtualGamePadButton Button);

	UPROPERTY(BlueprintReadOnly)
	bool bIsPossessedByPlayer1;
	
	// Called when the game starts or when spawned
	virtual void BeginPlay() override;

private:
	// reference to the game mode to get global game constants
	AKOFDefaultGameMode* GameMode;
	
public:	
	// Called every frame
	virtual void Tick(float DeltaTime) override;

};
